// vim: set foldmethod=marker :
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/twi.h>
#include <stdint.h>
#include "i2c/i2c.h"


#define I2C_IDLE	0
#define I2C_MTX		1
#define I2C_MRX		2
#define I2C_STX		3
#define I2C_SRX		4


#ifdef I2C_MASTER
static uint8_t* masterBuf;
static uint8_t masterBufSize;
static uint8_t masterBufIdx;
static uint8_t masterBufCount;
static uint8_t slarw;
static uint8_t error = I2C_E_OK;
static volatile uint8_t state = I2C_IDLE;
#endif

#ifdef I2C_SLAVE
static uint8_t slave_addr = 0;
static i2cRXHandler rxHandler = 0;
static i2cRequestHandler requestHandler = 0;
static uint8_t* slaveBuf;
static uint8_t slaveBufSize;
static uint8_t slaveBufIdx;
static uint8_t slaveBufCount;
static volatile uint8_t inCallback = 0;
#endif

void i2c_init(void)/*{{{*/
{
	TWBR = 7;
	TWSR = 1;
#ifdef I2C_SLAVE
	TWAR = slave_addr << 1;
#else
	TWAR = 0;
#endif
	TWCR = 
			_BV(TWINT)
		|	_BV(TWEA)
		|	_BV(TWEN)
		|	_BV(TWIE)
		;

}/*}}}*/

#ifdef I2C_SLAVE
void i2c_init_slave(uint8_t addr)/*{{{*/
{
	slave_addr = addr;
	i2c_init();
}/*}}}*/
#endif

#ifdef I2C_MASTER
void i2c_setMasterBuf(uint8_t* buf, uint8_t size)/*{{{*/
{
	masterBuf = buf;
	masterBufSize = size;
}
/*}}}*/
#endif

#ifdef I2C_SLAVE
void i2c_setSlaveBuf(uint8_t* buf, uint8_t size)/*{{{*/
{
	slaveBuf = buf;
	slaveBufSize = size;
}/*}}}*/
#endif

static void reply(uint8_t ack)/*{{{*/
{
	if(ack){
		TWCR =
				_BV(TWEN)
			|	_BV(TWIE)
			|	_BV(TWINT)
			|	_BV(TWEA)
			;
	}else{
		TWCR =
				_BV(TWEN)
			|	_BV(TWIE)
			|	_BV(TWINT);
	}
}/*}}}*/

#ifdef I2C_MASTER
static void send_stop(void)/*{{{*/
{
	TWCR = 
			_BV(TWEN)
		|	_BV(TWIE)
		|	_BV(TWEA)
		|	_BV(TWINT)
		|	_BV(TWSTO)
		;
	
	//while(TWCR & _BV(TWSTO));
	state = I2C_IDLE;
}/*}}}*/

#endif
static void release(void)/*{{{*/
{
	TWCR =
			_BV(TWEN)
		|	_BV(TWIE)
		|	_BV(TWEA)
		|	_BV(TWINT)
		;
#ifdef I2C_MASTER
	state = I2C_IDLE;
#endif
}/*}}}*/

#ifdef I2C_SLAVE
void i2c_onRX(i2cRXHandler callback)/*{{{*/
{
	rxHandler = callback;
}/*}}}*/

void i2c_onRequest(i2cRequestHandler callback)/*{{{*/
{
	requestHandler = callback;
}/*}}}*/
#endif

ISR(SIG_2WIRE_SERIAL)/*{{{*/
{
	uint8_t status = TWSR & 0xF8;

	switch (status) {

#ifdef I2C_SLAVE
		// slave receiver			/*{{{*/

		case TW_SR_ARB_LOST_SLA_ACK:
		case TW_SR_SLA_ACK:
#ifdef I2C_MASTER
			state = I2C_SRX;
#endif
			slaveBufCount = 0;
			if (slaveBufSize) {
				reply(1);
			} else {
				reply(0);
			}
#ifdef I2C_MASTER
			if (status == TW_SR_ARB_LOST_SLA_ACK) {
				error = I2C_E_ARB;
			}
#endif
			break;
			
		case TW_SR_DATA_ACK:
			if (slaveBufCount < slaveBufSize) {
				slaveBuf[slaveBufCount] = TWDR;
				slaveBufCount++;
			}
			if (slaveBufCount < slaveBufSize) {
				reply(1);
			} else {
				reply(0);
			}
			break;

		case TW_SR_DATA_NACK:
			reply(1);
			break;

		case TW_SR_STOP:
			if (rxHandler) {
				inCallback = 1;
				slaveBufIdx = 0;
				rxHandler();
				inCallback = 0;
			}
#ifdef I2C_MASTER
			state = I2C_IDLE;
#endif
			reply(1);
			break;

		/*}}}*/

		// slave transmitter		/*{{{*/
		case TW_ST_SLA_ACK:
		case TW_ST_ARB_LOST_SLA_ACK:
#ifdef I2C_MASTER
			state = I2C_STX;
#endif
			slaveBufIdx = 0;
			slaveBufCount = 0;
			if (requestHandler) {
				inCallback = 1;
				requestHandler();
				inCallback = 0;
			}
			if (!slaveBufCount) {
				*slaveBuf = 0;
				slaveBufCount = 1;
			}

#ifdef I2C_MASTER
			if (status == TW_ST_ARB_LOST_SLA_ACK) {
				error = I2C_E_ARB;
			}
#endif
			// fallback
		case TW_ST_DATA_ACK:
			TWDR = slaveBuf[slaveBufIdx];
			slaveBufIdx++;
			if (slaveBufIdx < slaveBufCount) {
				reply(1);
			} else {
				reply(0);
			}
			break;

		case TW_ST_LAST_DATA:
		case TW_ST_DATA_NACK:
			reply(1);

#ifdef I2C_MASTER
			state = I2C_IDLE;
#endif
			break;
		/*}}}*/
#endif

#ifdef I2C_MASTER
		// master transmitter/*{{{*/
		case TW_MT_SLA_ACK:
		case TW_MT_DATA_ACK:
			if (masterBufIdx < masterBufCount) {
				TWDR = masterBuf[masterBufIdx];
				masterBufIdx++;
				reply(1);
			} else {
				send_stop();
			}
			break;

		case TW_MT_SLA_NACK:
			error = I2C_E_ADDR_NACK;
			send_stop();
			break;
		case TW_MT_DATA_NACK:
			error = I2C_E_DATA_NACK;
			send_stop();
			break;

		case TW_MT_ARB_LOST:
			error = I2C_E_ARB;
			release();
			break;
		/*}}}*/

		//master receiver/*{{{*/

		case TW_MR_DATA_ACK:
			if (masterBufCount < masterBufSize) {
				masterBuf[masterBufCount] = TWDR;
				masterBufCount++;
			} else {
				error = I2C_E_BUFSIZE;
			}
			// fallback
		case TW_MR_SLA_ACK:
			if (masterBufCount < masterBufSize) {
				reply(1);
			} else {
				reply(0);
			}
			break;
		case TW_MR_DATA_NACK:
			if (masterBufCount < masterBufSize) {
				masterBuf[masterBufCount] = TWDR;
				masterBufCount++;
			} else {
				error = I2C_E_BUFSIZE;
			}
			// fallback
		case TW_MR_SLA_NACK:
			send_stop();
			break;
		/*}}}*/

		// master/*{{{*/
		case TW_START:
		case TW_REP_START:
			TWDR = slarw;
			reply(1);
			break;
		/*}}}*/
#endif
		// common/*{{{*/
		case TW_NO_INFO:
			break;
		case TW_BUS_ERROR:
#ifdef I2C_MASTER
		  error = I2C_E_BUS;
		  send_stop();
#else
		  release();
#endif
		  break;

		/*}}}*/

	}
}/*}}}*/

#ifdef I2C_MASTER
uint8_t i2c_getMasterBufSize(void)/*{{{*/
{
	return masterBufSize;
}/*}}}*/
#endif

#ifdef I2C_SLAVE
uint8_t i2c_getSlaveBufSize(void)/*{{{*/
{
	return slaveBufSize;
}/*}}}*/
#endif

#ifdef I2C_MASTER
void i2c_startTransmission(void)/*{{{*/
{
	masterBufCount = 0;
}/*}}}*/
#endif

void i2c_send(uint8_t data)/*{{{*/
{
#ifdef I2C_SLAVE
#ifdef I2C_MASTER
	if (inCallback) {
#endif
		if (slaveBufCount < slaveBufSize) {
			slaveBuf[slaveBufCount] = data;
			slaveBufCount++;
		}
#ifdef I2C_MASTER
	} else {
#endif
#endif
#ifdef I2C_MASTER
		if (masterBufCount < masterBufSize) {
			masterBuf[masterBufCount] = data;
			masterBufCount++;
		}
#ifdef I2C_SLAVE
	}
#endif
#endif
}/*}}}*/

#ifdef I2C_MASTER
uint8_t i2c_endTransmission(uint8_t addr)/*{{{*/
{
	slarw = TW_WRITE | (addr << 1);
	while (state != I2C_IDLE);

	state = I2C_MTX;
	masterBufIdx = 0;
	error = I2C_E_OK;
	TWCR = 
			_BV(TWINT)
		|	_BV(TWSTA)
		|	_BV(TWEN)
		|	_BV(TWIE)
		;
	while (state == I2C_MTX);
	
	return error;
}/*}}}*/

uint8_t i2c_request(uint8_t addr)/*{{{*/
{
	slarw = TW_READ | (addr << 1);
	while (state != I2C_IDLE);

	state = I2C_MRX;
	masterBufIdx = 0;
	masterBufCount = 0;
	error = I2C_E_OK;
	TWCR = 
			_BV(TWINT)
		|	_BV(TWSTA)
		|	_BV(TWEN)
		|	_BV(TWIE)
		;
	while (state == I2C_MRX);

	return error;
}/*}}}*/
#endif

uint8_t i2c_receive(void)/*{{{*/
{
#ifdef I2C_SLAVE
#ifdef I2C_MASTER
	if (inCallback) {
#endif
		if (slaveBufIdx < slaveBufCount) {
			return slaveBuf[slaveBufIdx++];
		} else {
			return 0;
		}
#ifdef I2C_MASTER
	} else {
#endif
#endif
#ifdef I2C_MASTER
		if (masterBufIdx < masterBufCount) {
			return masterBuf[masterBufIdx++];
		} else {
			return 0;
		}
#ifdef I2C_SLAVE
	}
#endif
#endif
}/*}}}*/

uint8_t i2c_available(void)/*{{{*/
{

#ifdef I2C_SLAVE
#ifdef I2C_MASTER
	if (inCallback) {
#endif
		return slaveBufCount - slaveBufIdx;
#ifdef I2C_MASTER
	}
#endif
#endif

#ifdef I2C_MASTER
#ifdef I2C_SLAVE
	if (!inCallback) {
#endif
		return masterBufCount - masterBufIdx;
#ifdef I2C_SLAVE
	}
#endif
#endif
	return 0;
}/*}}}*/

